#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include <sys/time.h>
#include <stdlib.h>
#include "llheap.h"
#include "graph.h"

 // Initialize a graph
Graph *init_graph (int gid, int num_nodes, int num_edges)
  {
    Graph *g = new Graph;
    g->id = gid;
    g->num_nodes = num_nodes;
    g->num_edges = num_edges;
    g->curr_num_nodes = 0;
    g->curr_num_edges = 0;
    g->node_list = new Node* [num_nodes];
    g->edge_list = new Edge* [num_edges];

    g->src = NULL;

    return g;
  }

 // Init a node
void init_node (Node *n, int index, int id, int xpos, int ypos)
  {
    assert(n != NULL);
    n->index = index;
    n->id = id;
    n->xpos = xpos;
    n->ypos = ypos;
    n->adj_count = 0;
    n->curr_adj_count = 0;
    n->ssp_computed = false;
    n->has_member = false;
    return;
  }

 // Init an edge
void init_edge (Edge *e, int index, int id, Node *node1, Node *node2, int wt)
  {
    assert(e != NULL);
    e->index = index;
    e->id = id;
    e->node1 = node1;
    e->node2 = node2;
    e->wt = wt;
    return;
  }

 // Add a node to a graph, with the given node id
void add_node (Graph *g, int index, int node_id, int xpos, int ypos)
  {
    assert(g != NULL);
    assert(g->curr_num_nodes < g->num_nodes);

    g->node_list[g->curr_num_nodes] = new Node;
    init_node(g->node_list[g->curr_num_nodes],index,node_id,xpos,ypos);
    g->curr_num_nodes ++;
    return;
  }

 // Returns pointer to the node structure with the matching node id.
 // No node in the node_list of the graph should be NULL.
Node *locate_node (Graph *g, int node_id)
  {
    assert(g != NULL);
    for (int i = 0; i < g->num_nodes; i++)
	if (g->node_list[i]->id == node_id)
		return g->node_list[i];
    return NULL;
  }

 // This routine should be called after all the nodes of the graph has
 // structures malloc-ed for them.
 // In the process also calculates the size of the adj list of each
 // node.
void add_edge (Graph *g, int index, int edge_id, int node1_id, int node2_id, int wt)
  {
    Node *node1, *node2;

    assert(g != NULL);
    assert(g->curr_num_edges < g->num_edges);
    g->edge_list[g->curr_num_edges] = new Edge;
    node1 = locate_node(g,node1_id);
    node2 = locate_node(g,node2_id);

    assert((node1 != NULL) && (node2 != NULL));

    node1->adj_count ++;
    node2->adj_count ++;
    init_edge(g->edge_list[g->curr_num_edges],index,edge_id,node1,node2,wt);
    g->curr_num_edges ++;
    return;
  }

 // Assign adj list for nodes -- puts the relevant Edge* in the adj_list of
 // the nodes.
void assign_adj_list (Graph *g)
  {
    assert (g != NULL);
    for (int i = 0; i < g->num_nodes; i++)
      {
	Node *n = g->node_list[i];
	n->adj_list = new Edge* [n->adj_count];
      }

    for (int j = 0; j < g->num_edges; j++)
      {
	Edge *e = g->edge_list[j];
	e->node1->adj_list[e->node1->curr_adj_count] = e;
	e->node2->adj_list[e->node2->curr_adj_count] = e;
	e->node1->curr_adj_count ++;
	e->node2->curr_adj_count ++;
      }

    return;
  }

// Reads the graph from the file named.
Graph* read_graph (char *file_name, int graph_id)
  {
    char str[MAX_STR_LEN];
    FILE *fp = fopen(file_name,"r");
    int num_nodes, num_edges;
    Graph *g;
    int node_id, node_xpos, node_ypos;
    int node1_id, node2_id, wt;

    if (fp == NULL)
      {
	fprintf (stderr,"Could not open file : %s\n", file_name);
	return NULL;
      }

    fgets(str,MAX_STR_LEN,fp);// This is a comment line.
    fgets(str,MAX_STR_LEN,fp);
    sscanf(str,"%d %d",&num_nodes, &num_edges);
    num_edges /= 2; // gt-itm graphs gives the sum of degrees, which is
                    // which is twice the sum of all the edges
    fgets(str,MAX_STR_LEN,fp); // Blank line
    fgets(str,MAX_STR_LEN,fp); // Comment line

    g = init_graph(graph_id,num_nodes,num_edges);
    for (int i = 0; i < num_nodes; i++)
      {
	fgets(str,MAX_STR_LEN,fp);
	sscanf(str,"%d %*s %d %d",&node_id,&node_xpos,&node_ypos);
	add_node(g,i,node_id,node_xpos,node_ypos);
      }

    fgets(str,MAX_STR_LEN,fp); // Blank line
    fgets(str,MAX_STR_LEN,fp); // Comment line

    for (int j = 0; j < num_edges; j++)
      {
	fgets(str,MAX_STR_LEN,fp);
	sscanf(str,"%d %d %d",&node1_id,&node2_id,&wt);
	add_edge(g,j,j,node1_id,node2_id,wt);
      }
    assign_adj_list(g);

    fprintf (stdout, "Created the graph from file : %s\n",file_name);
    fclose(fp);
    return g;
  }

// Output a graph specs into the filename specified
void output_graph (Graph *g, char *file_name)
  {
    FILE *fp = fopen(file_name,"w");
    if (fp == NULL)
      {
	fprintf(stderr,"Could not open file : %s\n",file_name);
	return;
      }

    fprintf(fp,"GRAPH #nodes #edges\n");
    fprintf(fp,"%d %d\n\n",g->num_nodes,g->num_edges);

    fprintf(fp,"VERTICES (index xpos ypos neighbor-count : dest-nodes)\n");
    for (int i = 0; i < g->num_nodes; i++)
      {
       Node *n = g->node_list[i];
       fprintf(fp,"%d %d %d %d : ",n->id,n->xpos,n->ypos,n->adj_count);
       for (int k = 0; k < n->adj_count; k++)
	 {
	   if (n->adj_list[k]->node1 == n)
	     fprintf(fp,"%d ",n->adj_list[k]->node2->id);
	   else
	     fprintf(fp,"%d ",n->adj_list[k]->node1->id);
	 }
       fprintf(fp,"\n");
      }

    fprintf(fp,"\nEDGES (from-node to-node length)\n");
    for (int j = 0; j < g->num_edges; j++)
      {
	Edge *e = g->edge_list[j];
	fprintf (fp,"%d %d %d\n",e->node1->id,e->node2->id,e->wt);
      }

    fclose(fp);
    fprintf (stdout, "Wrote the graph to file : %s\n",file_name);
    return;
 }

void write_graph_in_myns_format (Graph *g, char * filename) {

  FILE *fp = fopen(filename,"w");
  if (fp == NULL)
    {
     fprintf(stderr,"Could not open file : %s\n",filename);
     return;
    }

  fprintf (fp,"nodecount %d;\n\n", g->num_nodes);
  for (int i = 0; i < g->num_edges; i++) {

    fprintf (fp,"edge %d %d %8.3f;\n", g->edge_list[i]->node1->id,  g->edge_list[i]->node2->id, ((double)(g->edge_list[i]->wt)) / SCALE_FACTOR);
  }
  fclose(fp);
  return;
}

// Delete the graph
void delete_graph (Graph *g)
{
  for (int i = 0; i < g->num_nodes; i++)
    {
      Node *n = g->node_list[i];
      if (n->ssp_computed)
	delete_ssp_tree(g,n);
      delete [] n->adj_list;
      delete n;
    }

  for (int j = 0; j < g->num_edges; j++)
    delete g->edge_list[j];

  delete g->node_list;
  delete g->edge_list;

  fprintf (stdout, "Deleted the graph : %d\n",g->id);
  delete g;
  return;
}
      
// Initialize for shortest path algo rooted at a source, s
void init_shortest_path (Graph *g, Node *s)
{
  assert(g != NULL);

  delete_ssp_tree(g,s);

  for (int i = 0; i < g->num_nodes; i++)
    {
      Node *n = g->node_list[i];
      if (i == s->index)
	s->ssp_d = 0;
      else
	n->ssp_d = INT_MAX/2;
      n->ssp_parent = NULL;
      n->ssp_num_children = 0;
      n->curr_ssp_num_children = 0;
    }
  return;
}

// Relax routine of Cormen et al book (pg. 520)
void shortest_path_relax (Node *u, Node*v, Edge *uv)
{
  //assert( ((uv->node1 == u) && (uv->node2 == v)) ||
  //	  ((uv->node2 == u) && (uv->node1 == v)) );

  // Changing the following from path weights to hop count
  // ie. uv->wt == 1
  //if (v->ssp_d > (u->ssp_d + uv->wt))
  if (v->ssp_d > (u->ssp_d + 1))
    {
      //v->ssp_d = u->ssp_d + uv->wt;
      v->ssp_d = u->ssp_d + 1;
      v->ssp_parent = u;
    }
  return;
}

// Perform the Dijkstra's single source shortest path algo
// Cormen et al. Page 527
void dijkstra_ssp (Graph *g, Node *s)
{
  if (s->ssp_computed)
    return;
  init_shortest_path(g,s);

  LLHeap *h;
  // Position *Pos = new Position [g->num_nodes];

  h = init_heap(g->num_nodes);
  for (int i = 0; i < g->num_nodes; i++)
      insert(h,g->node_list[i]->ssp_d);
      
  while (!is_empty(h))
    {
      int min_sspd;
      int min_index = delete_min(h,min_sspd);
      assert (min_index >= 0);
      for (int j = 0; j < g->node_list[min_index]->adj_count; j++)
	{
	  Edge *e = g->node_list[min_index]->adj_list[j];
	  Node *this_node, *other_node;
	  if (e->node1 == g->node_list[min_index])
	    {
	      this_node = e->node1;
	      other_node = e->node2;
	    }
	  else
	    {
	      this_node = e->node2;
	      other_node = e->node1;
	    }
	  if (heap_check_valid(h,other_node->index))
	    {
	      shortest_path_relax(this_node,other_node,e);
	      change_key(h,other_node->index,other_node->ssp_d);
	    }
	}
    }
  delete_heap(h);

  s->ssp_tree = new SSP [g->num_nodes];

  // Copy the parent info into the SSP structure and also compute the
  // number of children of each node.
  for (int k = 0; k < g->num_nodes; k++)
    {
      s->ssp_tree[k].level = -1; //Init the level field
      s->ssp_tree[k].parent = g->node_list[k]->ssp_parent;
      if (g->node_list[k]->ssp_parent != NULL)
	g->node_list[k]->ssp_parent->ssp_num_children ++;
    }

  for (int m = 0; m < g->num_nodes; m++)
    {
      Node *n = g->node_list[m];
      n->ssp_children = new Node* [n->ssp_num_children];
    }

  for (int q = 0; q < g->num_nodes; q++)
    {
      Node *sp = g->node_list[q]->ssp_parent;
      if (sp != NULL)
	{
	  sp->ssp_children[sp->curr_ssp_num_children] = g->node_list[q];
	  sp->curr_ssp_num_children ++;
	}
    }

  for (int p = 0; p < g->num_nodes; p++)
    {
      s->ssp_tree[p].children = g->node_list[p]->ssp_children;
      s->ssp_tree[p].num_children = g->node_list[p]->ssp_num_children;
    }

  assign_ssp_level(g,s);
  s->ssp_computed = true;
  fprintf (stdout,"Created the shortest path tree sourced at : %d\n", s->id);
  return;
}

// Performs the Bellman Ford single source shortest path algo
// Uses the algo mentioned in Cormen et al.
void bellman_ford_ssp (Graph *g, Node *s)
{

  if (s->ssp_computed)
    return;

  init_shortest_path(g,s);
  for (int i = 0; i < g->num_nodes - 1; i++)
    for (int j = 0; j < g->num_edges; j++)
      {
       Edge *e = g->edge_list[j];
       shortest_path_relax(e->node1,e->node2,e);
       shortest_path_relax(e->node2,e->node1,e);
      }

  s->ssp_tree = new SSP [g->num_nodes];

  // Copy the parent info into the SSP structure and also compute the
  // number of children of each node.
  for (int k = 0; k < g->num_nodes; k++)
    {
      s->ssp_tree[k].level = -1; //Init the level field
      s->ssp_tree[k].parent = g->node_list[k]->ssp_parent;
      if (g->node_list[k]->ssp_parent != NULL)
	g->node_list[k]->ssp_parent->ssp_num_children ++;
    }

  for (int m = 0; m < g->num_nodes; m++)
    {
      Node *n = g->node_list[m];
      n->ssp_children = new Node* [n->ssp_num_children];
    }

  for (int q = 0; q < g->num_nodes; q++)
    {
      Node *sp = g->node_list[q]->ssp_parent;
      if (sp != NULL)
	{
	  sp->ssp_children[sp->curr_ssp_num_children] = g->node_list[q];
	  sp->curr_ssp_num_children ++;
	}
    }

  for (int p = 0; p < g->num_nodes; p++)
    {
      s->ssp_tree[p].children = g->node_list[p]->ssp_children;
      s->ssp_tree[p].num_children = g->node_list[p]->ssp_num_children;
    }

  assign_ssp_level(g,s);
  s->ssp_computed = true;
  fprintf (stdout,"Created the shortest path tree sourced at : %d\n", s->id);
  return;
}

// Assigns the SSP tree levels for all nodes
void assign_ssp_level (Graph *g, Node *s)
{
 for (int i = 0; i < g->num_nodes; i++)
   assign_ssp_level_from_parent(g,s,g->node_list[i]);
 return;
}

// Assigns the SSP tree levels in the tree, by first assigning the SSP tree level
// for the parent. s : is the source node, and n is the node whose level is being
// assigned.
void assign_ssp_level_from_parent (Graph *g, Node *s, Node *n)
{
  if (s->ssp_tree[n->index].level > 0) // This means the level is already assigned.
    return;

  if (s == n)
    {
      n->ssp_tree[n->index].level = 0;
      return;
    }
  Node *pt_of_n = s->ssp_tree[n->index].parent;
  assert(pt_of_n != NULL);
  if (s->ssp_tree[pt_of_n->index].level < 0) // This would be true initally.
    assign_ssp_level_from_parent(g,s,pt_of_n);
  s->ssp_tree[n->index].level = s->ssp_tree[pt_of_n->index].level + 1;

  return;
}

  
// Delete the SSP tree
void delete_ssp_tree (Graph *g, Node *s)
{
  if (!(s->ssp_computed))
    return;
  for (int i = 0; i < g->num_nodes; i++)
    delete [] s->ssp_tree[i].children;
  delete [] s->ssp_tree;
  s->ssp_computed = false;
  fprintf (stdout, "SSP tree deleted for node %d\n", s->id);
  return;
}

// Output the SSP tree
void out_ssp_tree (Graph *g, Node *s)
{
 if (!(s->ssp_computed))
   {
     fprintf(stdout,"SSP not computed\n");
     return;
   }

 fprintf (stdout, "SSP Tree of %d\n", s->id);
 for (int i = 0; i < g->num_nodes; i++)
   {
     fprintf (stdout, "Node %d :: Level %d :: ", g->node_list[i]->id,s->ssp_tree[i].level);
     if (g->node_list[i]->has_member)
       fprintf (stdout,"Member %d :: ",g->node_list[i]->member_id);
     else
       fprintf (stdout,"No Member :: ");

     if (s->ssp_tree[i].parent != NULL)
       fprintf (stdout, "Parent  %d :: ", s->ssp_tree[i].parent->id);
     else
       fprintf (stdout, "Parent  NULL :: ");
     fprintf (stdout, "Num Children %d :: Children ", s->ssp_tree[i].num_children);
     for (int j = 0; j < s->ssp_tree[i].num_children; j++)
       fprintf (stdout,"%d ",s->ssp_tree[i].children[j]->id);
     fprintf (stdout,"\n");
   }
 fprintf (stdout, "Tree display complete\n");
 return;
}

// Arbitrarily assigns cnt mcast group members to the nodes.
void assign_random_members (Graph *g, int cnt)
{
  int num_free_nodes; // Nodes that dont have members assigned
  int locate_free_index; // Finds the next index position of the nodes as required
  // by new_member_index
  int new_member_index; // The next member position (ignoring the ones that are
  // already been assigned.

  if (cnt > g->num_nodes)
    {
      fprintf (stderr, "Too many members : %d, should be upto %d\n",cnt,g->num_nodes);
      return;
    }
  num_free_nodes = g->num_nodes;
  for (int i = 0; i < cnt; i++)
    {
      new_member_index = rand() % num_free_nodes;
      locate_free_index = 0;
      for (int j = 0; j < g->num_nodes; j++)
	{
	  Node *n = g->node_list[j];
	  if (!(n->has_member))
	    {
	     if (locate_free_index == new_member_index)
	      {
	       n->has_member = true;
	       n->member_id = i;
	       num_free_nodes --;
	       break;
	      }
	     else
	      locate_free_index ++;
	    }
	}
    }
  return;
}

// Assign the members as specified.
void assign_specific_members_and_source (Graph *g, char *file_name)
{
  FILE *fp = fopen(file_name,"r");
  char str[MAX_STR_LEN];
  int cnt;
  int node_id, member_id;

  if (fp == NULL)
    {
      fprintf (stderr,"Could not open file : %s\n",file_name);
      return;
    }

  fgets(str,MAX_STR_LEN,fp); // Comment line
  fgets(str,MAX_STR_LEN,fp); // Has the number of members
  sscanf (str,"%d",&cnt);
  fgets(str,MAX_STR_LEN,fp); // Comment line
  
  for (int i = 0; i < cnt; i++)
    {
      fgets(str,MAX_STR_LEN,fp); // Has <node_id member_id>
      sscanf(str,"%d %d",&node_id, &member_id);
      for (int j = 0; j < g->num_nodes; j++)
	{
	  Node *n = g->node_list[j];
	  if (n->id == node_id)
	    {
	      assert(!(n->has_member));
	      n->has_member = true;
	      n->member_id = member_id;
	      break;
	    }
	}
    }

  fgets(str,MAX_STR_LEN,fp); // Blank line
  fgets(str,MAX_STR_LEN,fp); // Comment line
  fgets(str,MAX_STR_LEN,fp);

  sscanf(str,"%d",&node_id); // This contains the source
  choose_specific_source(g,node_id);

  fclose(fp);
  return;
}

// Outputs the group members in the given file_name
void output_group_members (Graph *g, char *file_name)
{
  FILE *fp = fopen(file_name,"w");
  int node_id, member_id;
  int count = 0;

  if (fp == NULL)
    {
      fprintf (stderr,"Could not open file : %s\n",file_name);
      return;
    }

  fprintf (fp,"Multicast group members\nNode-id Member-id\n");
  for (int i = 0; i < g->num_nodes; i++)
    {
      Node *n = g->node_list[i];
      if (n->has_member)
	{
	  count ++;
	  fprintf (fp,"%d %d\n",n->id,n->member_id);
	}
    }

  fprintf (fp,"Number of members : %d\n",count);
  fprintf (fp,"\n Source node : %d\n",g->src->id);
  fclose(fp);
  return;
}
 
// Chooses a source arbitarily from the mcast members, of which there are cnt.
Node *choose_random_source (Graph *g, int cnt)
{
 int ndx = rand() % cnt;
 int pos = 0;

 for (int i = 0; i < g->num_nodes; i++)
   {
     Node *n = g->node_list[i];
     if (n->has_member)
       {
        if (pos == ndx)
	  {
	    g->src = n;
	    return n;
	  }
	else
	  pos ++;
       }
   }
 return NULL;
}

// Chooses a specific node to be the source.
Node *choose_specific_source (Graph *g, int node_id)
{
 for (int i = 0; i < g->num_nodes; i++)
   {
     Node *n = g->node_list[i];
     if (n->id == node_id)
       if (n->has_member == true)
         {
	  g->src = n;
	  return n;
	 }
   }
 return NULL;
}

// Computes the SSP for all members in the graph
void compute_ssp_members (Graph *g)
{
  for (int i = 0; i < g->num_nodes; i++)
    if (g->node_list[i]->has_member)
      dijkstra_ssp(g,g->node_list[i]);
      //bellman_ford_ssp(g,g->node_list[i]);
  return;
}

// Returns the node which is the member parent of the given node n.
Node *locate_member_parent (Graph *g, Node *n)
{
 assert(n->has_member);
 assert (g->src->ssp_computed);

 // Start with the immediate parent.
 Node *p = g->src->ssp_tree[n->index].parent;

 while (p != NULL)
   {
     // If this ancestor has a member, then this is the first such
     // ancestor. So it must be the parent.
     if (p->has_member)
       return p;
     // Else go to the next higher ancestor and repeat.
     Node *new_p = g->src->ssp_tree[p->index].parent;
     if (new_p == NULL)
       assert(p == g->src);
     p = new_p;
   }
 return g->src;
}

// Outputs to stdout, the shortest path from x -> y as per SSP of x
void output_x_to_y_path (Graph *g, Node *x, Node *y)
{
  int *ancestor_list;

  if (!(x->ssp_computed))
      {
	fprintf (stdout, "SSP of node %d was not computed\n",x->id);
	bellman_ford_ssp(g,x);
      }

  int i = x->ssp_tree[y->index].level;
  ancestor_list = new int [ i + 1];
  ancestor_list[i] = y->id;


  Node *p = x->ssp_tree[y->index].parent;

  fprintf (stdout,"Output of path from %d to %d\n",x->id,y->id);
  while (p != NULL)
    {
      i--;
      if (i < 0)
	fprintf (stderr,"Problem : level and number of ancestors dont match\n");
      else
	ancestor_list[i] = p->id;
      Node *new_p = x->ssp_tree[p->index].parent;
      if (new_p == NULL)
	assert(p == x);
      p = new_p;
    }

  if (i > 0)
    fprintf (stdout,"Valid ancestors start from pos : %d\n",i);
  for (int j = 0; j < x->ssp_tree[y->index].level; j++)
      fprintf (stdout, "%d=> %d\n", j, ancestor_list[j]);

  return;
}
  

// Outputs the possibility of errors
void output_parent_errors (Graph *g, char *file_name)
{
  FILE *fp = fopen(file_name,"w");
  if (fp == NULL)
    {
      fprintf (stderr,"Could not open file : %s\n", file_name);
      return;
    }

  Node *s = g->src; // The source node.
  int count = 0;
  int total_members_lower_level_mistake = 0;
  int total_members_same_level_mistake = 0;

  for (int i = 0; i < g->num_nodes; i++)
    {
      Node *n = g->node_list[i];
      if (n == s)
	continue; // No need to check for the source
      if (!(n->has_member))
	continue;

      count ++;
      // For each member compute the error
      assert(n->ssp_computed);
      // Here, n stands for this node, p is the correct member parent
      // and s is source
      int s_n_ttl = s->ssp_tree[i].level; // TTL(s,n)
      Node *p = locate_member_parent(g,n);
      int s_p_ttl = s->ssp_tree[p->index].level; // TTL(s,p)

      assert(p->ssp_computed); // Need this for TTL(p,n)
      int p_n_ttl = p->ssp_tree[n->index].level; // TTL(p,n)

      if (s_p_ttl + p_n_ttl != s_n_ttl) // This is an error, the assert fails
	{
	  fprintf (stderr,"Assert going to fail for node %d, parent %d and source %d\n", n->id, p->id, s->id);
	  fprintf (stderr, "TTL(s,p) : %d, TTL(p,n) : %d, TTL(s,n) : %d\n", s_p_ttl, p_n_ttl, s_n_ttl);
	  output_x_to_y_path(g,s,p);
	  output_x_to_y_path(g,p,n);
	  output_x_to_y_path(g,s,n);
	}
      assert(s_p_ttl + p_n_ttl == s_n_ttl);

      fprintf (fp, "Node %d : [P - %d] : ",n->id, p->id);
      int same_level_mistake = 0; // Mistake with a member at same level
      // as parent
      int lower_level_mistake = 0; // Mistake with a member closer than parent
      for (int j = 0; j < g->num_nodes; j++)
	{
	  Node *t = g->node_list[j];
	  if (!(t->has_member))
	    continue;
	  int s_t_ttl = s->ssp_tree[t->index].level;
	  if (s_t_ttl < s_p_ttl)
	    continue; // There is no mistake possible if this node is higher
	  // up in the tree than the parent

	  int t_n_ttl = t->ssp_tree[n->index].level;
	  if (s_t_ttl + t_n_ttl != s_n_ttl)
	    continue; // No mistake possible here either

	  assert(s_t_ttl + t_n_ttl == s_n_ttl); // Can't have less
	  if ((s_t_ttl == s_p_ttl) && (t != p)) 
	    {
	      // Then this is a same level mistake (ie. same level as parent)
	      same_level_mistake ++;
	      fprintf (fp,"(S - %d) ",t->id);
	    }

	  if ((s_t_ttl > s_p_ttl) && (t != n))
	    {
	      assert (t != n);
	      lower_level_mistake ++;
	      fprintf (fp, "(L - %d) ",t->id);
	    }
	}
      fprintf (fp,"\n\tSame level : %d\tLower level : %d\n", same_level_mistake, lower_level_mistake);

      // Compute member count for mistakes for the entire tree
      if (lower_level_mistake > 0)
	total_members_lower_level_mistake ++;
      else if (same_level_mistake > 0)
	total_members_same_level_mistake ++;
	
    }

  fprintf (fp,"Num nodes : %d\tNum members : %d\n",g->num_nodes, count);
  fprintf (fp, "Members lower level mistake : %d\n",total_members_lower_level_mistake);
  fprintf (fp, "Members same level mistake : %d (does not include the members who have already made the lower level mistake)\n",total_members_same_level_mistake);
  fclose(fp);
  return;
}

int main(int argc, char **argv)
{
  Graph *g;
  struct timeval tp;
  struct timezone tzp;

  // Init the random number generator
  gettimeofday(&tp,&tzp);
  srand((int)(tp.tv_usec));

 if (argc != 4) 
   {
     fprintf(stderr,"Not enough arguments\n");
     fprintf(stderr, "Format : %s <graph_specs_file> <test_output_file> <myns_output_file>\n",argv[0]);
     fprintf (stderr, "\t graph_specs_file : GT-ITM .alt file for the topology\n");
     fprintf (stderr, "\t test_output_file : Echoes the .alt file\n");
     fprintf (stderr, "\t myns_output_file : Output in myns format\n");
     return -1;
   }

 g = read_graph(argv[1],1);
 if (g == NULL)
   {
     fprintf (stderr, "Could not create graph\n");
     return -1;
   }
 if (strcmp(argv[2],"-") != 0)
   output_graph(g,argv[2]);
 write_graph_in_myns_format(g,argv[3]);

 /*
 if (strcmp(argv[3],"NULL") != 0)
     assign_specific_members_and_source(g,argv[3]);
 else
   {
     int count;
     sscanf(argv[6],"%d",&count);
     assign_random_members(g,count);
     Node *src = choose_random_source(g,count);
     output_group_members(g,argv[4]);
     fprintf (stdout,"Chosen source node : %d\n",src->id);
   }

 dijkstra_ssp(g,g->node_list[0]);
 out_ssp_tree(g,g->node_list[0]);
 delete_ssp_tree(g,g->node_list[0]);
 return 1;
 compute_ssp_members(g);
 output_parent_errors(g,argv[5]);
 */

 delete_graph(g);
 return 1;
}

